package de.fhmuenster.masterthesis.generator;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.fhmuenster.masterthesis.generator.mockdata.ExternalServiceMock;
import de.fhmuenster.masterthesis.generator.mockdata.IsMock;
import de.fhmuenster.masterthesis.generator.testdata.IsPassingCheck;
import de.fhmuenster.masterthesis.generator.testdata.IsTeststep;
import de.fhmuenster.masterthesis.generator.testdata.IsTeststepWithInput;
import de.fhmuenster.masterthesis.generator.testdata.NotPassedCheck;
import de.fhmuenster.masterthesis.generator.testdata.PassedCheck;
import de.fhmuenster.masterthesis.generator.testdata.ProcessEndedCheck;
import de.fhmuenster.masterthesis.generator.testdata.ProcessStartedCheck;
import de.fhmuenster.masterthesis.generator.testdata.TestExternalTask;
import de.fhmuenster.masterthesis.generator.testdata.TestProgressCheck;
import de.fhmuenster.masterthesis.generator.testdata.TestStart;
import de.fhmuenster.masterthesis.generator.testdata.TestTask;
import de.fhmuenster.masterthesis.generator.testdata.Testcase;
import de.fhmuenster.masterthesis.generator.testdata.VariableValuePair;
import de.fhmuenster.masterthesis.generator.testdata.VariablesCheck;

public class BPMNTestcollectionCreator {

	private String processInstanceKey;

	private List<PassCheck> passCheckList;

	public BPMNTestcollectionCreator() {
	}

	public BPMNTestcollectionCreator withProcessIntanceKey(String processInstanceKey) {
		this.processInstanceKey = processInstanceKey;

		return this;
	}

	public Testcase build(BPMNGenTestcase bpmnTestcase) {
		createPassCheckMap(bpmnTestcase);

		Testcase testcase = new Testcase();
		testcase.setTestcaseName(bpmnTestcase.getName());

		List<BPMNGenFlowElement> flowElements = bpmnTestcase.getFlow().getFlowNodes();
		
		optimizePassCheckOrder(flowElements);

		List<List<BPMNGenFlowElement>> fragments = fragmentToActAndAssertBundles(flowElements);
		fragments.stream() //
				.flatMap(fragment -> createTeststeps(fragment, bpmnTestcase).stream()) //
				.forEach(teststep -> testcase.addTeststep(teststep)); //

		resetPassCheckMap();

		return testcase;
	}

	private List<List<BPMNGenFlowElement>> fragmentToActAndAssertBundles(List<BPMNGenFlowElement> flowElements) {
		List<List<BPMNGenFlowElement>> fragments = new ArrayList<>();
		List<BPMNGenFlowElement> fragmentElements = new ArrayList<>();

		for (BPMNGenFlowElement flowElement : flowElements) {
			if (fragmentCutNeeded(flowElement)) {
				fragments.add(fragmentElements);
				fragmentElements = new ArrayList<>();
			}

			fragmentElements.add(flowElement);
		}

		fragments.add(fragmentElements);

		return fragments;
	}

	private boolean fragmentCutNeeded(BPMNGenFlowElement flowElement) {
		if (flowElement.getType().equals(FlowElementType.USER_TASK) || //
				flowElement.getType().equals(FlowElementType.SERVICE_TASK_EXTERNAL)) {
			return true;
		}

		return false;
	}

	private List<IsTeststep> createTeststeps(List<BPMNGenFlowElement> fragment, BPMNGenTestcase testcase) {
		markFragmentFlowElementsAsShouldHavePassed(fragment);

		List<IsTeststep> teststeps = new ArrayList<>();
		boolean checkStarted = false;
		boolean shouldBeEnded = false;

		BPMNGenFlowElement startElement = fragment.get(0);
		BPMNGenFlowElement endElement = fragment.get(fragment.size() - 1);
		HashMap<String, HashMap<String, Object>> globalTaskVariableMap = testcase.getGlobalTaskVariableMap();
		Map<String, String> globalExternalTopicsMap = testcase.getGlobalExternalTopicsMap();
		if (startElement.getType().equals(FlowElementType.START_EVENT)) {
			checkStarted = true;

			TestStart testStart = createTestStart(startElement, testcase.getTaskVariableMap());
			teststeps.add(testStart);
		}

		if (startElement.getType().equals(FlowElementType.USER_TASK)) {
			TestTask testTask = createTestTask(startElement, testcase.getTaskVariableMap());
			teststeps.add(testTask);
		}
		
		if (startElement.getType().equals(FlowElementType.SERVICE_TASK_EXTERNAL)) {
			HashMap<String, Object> hardcodedVariableValues = globalTaskVariableMap.getOrDefault(startElement.getId(), new HashMap<>());
			TestExternalTask testExternalTask = createTestExternalTask(startElement, testcase.getMocks(), globalExternalTopicsMap, hardcodedVariableValues);
			teststeps.add(testExternalTask);
		}

		if (endElement.getType().equals(FlowElementType.END_EVENT)) {
			shouldBeEnded = true;
		}

		TestProgressCheck progressCheck = createTestProgressCheck(checkStarted, shouldBeEnded, testcase.getEndCheckVariableMap());
		teststeps.add(progressCheck);

		return teststeps;
	}

	private TestStart createTestStart(BPMNGenFlowElement startEvent,
			List<TaskVariables> taskVariables) {
		TestStart testStart = new TestStart(startEvent.getId());
		testStart.setProcessInstanceKey(processInstanceKey);
		addVariablesToTeststep(testStart, getVariablesAndPop(taskVariables, startEvent.getId()));

		return testStart;
	}

	private TestTask createTestTask(BPMNGenFlowElement userTask, List<TaskVariables> taskVariables) {
		TestTask testTask = new TestTask(userTask.getId());
		addVariablesToTeststep(testTask, getVariablesAndPop(taskVariables, userTask.getId()));

		return testTask;
	}
	
	private TestExternalTask createTestExternalTask(BPMNGenFlowElement externalServiceTask, List<IsMock> mocks, Map<String, String> globalExternalTopicsMap, HashMap<String, Object> hardcodedVariableValues) {
		String externalTopicName = globalExternalTopicsMap.get(externalServiceTask.getId());
		String externalMockName = getExternalMockName(mocks, externalTopicName, globalExternalTopicsMap);
		
		TestExternalTask testExternalTask = new TestExternalTask();
		testExternalTask.setName(externalServiceTask.getId());
		testExternalTask.setTopic(externalTopicName);
		testExternalTask.setExternalMockName(externalMockName);
		testExternalTask.setHardcodedVariableValues(hardcodedVariableValues);
		
		return testExternalTask;
	}
	
	private String getExternalMockName(List<IsMock> mocks, String externalTopicName, Map<String, String> globalExternalTopicsMap) {
		if(externalTopicName != null) {
			for (IsMock mock : mocks) {
				if(mock instanceof ExternalServiceMock && externalTopicName.equals(((ExternalServiceMock) mock).getExternalTopic())) {
					return mock.getName();
				}
			}
		}
		
		return null;
	}

	private TestProgressCheck createTestProgressCheck(boolean checkStarted, boolean shouldBeEnded, HashMap<String, VariableCheck> endVariableChecks) {
		TestProgressCheck progressCheck = new TestProgressCheck();

		if (checkStarted) {
			ProcessStartedCheck startedCheck = new ProcessStartedCheck(true);
			progressCheck.addProgressCheck(startedCheck);
		}

		passCheckList.stream() //
				.map(passCheck -> mapToPassingCheck(passCheck.getFlowElementId(), passCheck.getFlowElement().getType(), passCheck.isPassCheck())) //
				.forEach(passingCheck -> progressCheck.addProgressCheck(passingCheck)); //
		
		ProcessEndedCheck endedCheck = new ProcessEndedCheck(shouldBeEnded);
		progressCheck.addProgressCheck(endedCheck);
		
		//Needs to be last check
		if(shouldBeEnded) {
			if(!endVariableChecks.isEmpty()) {
				VariablesCheck variablesCheck = new VariablesCheck();
				endVariableChecks.entrySet().stream() //
					.filter(check -> check.getValue().getOperator().equals(VariableCheckOperator.EQUALS)) //
					.map(check -> new VariableValuePair(check.getKey(), check.getValue().getValue())) //
					.forEach(check -> variablesCheck.addVariableEqualsCheck(check)); //
				
				endVariableChecks.entrySet().stream() //
					.filter(check -> check.getValue().getOperator().equals(VariableCheckOperator.NOT_EQUALS)) //
					.map(check -> new VariableValuePair(check.getKey(), check.getValue().getValue())) //
					.forEach(check -> variablesCheck.addVariableNotEqualsCheck(check)); //
				
				progressCheck.addProgressCheck(variablesCheck);
			}
		}

		return progressCheck;
	}

	private void addVariablesToTeststep(IsTeststepWithInput teststep, HashMap<String, Object> variables) {
		if (variables != null) {
			variables.entrySet().forEach(entry -> teststep
					.addVariableKeyValue(NameProvider.getStaticFinalVariableName(entry.getKey()), entry.getValue()));
		}
	}

	private void createPassCheckMap(BPMNGenTestcase bpmnTestcase) {
		passCheckList = new ArrayList<>();

		List<BPMNGenFlowElement> flowElements = bpmnTestcase.getFlow().getFlowNodes();
		List<BPMNGenFlowElement> exclusiveFlowElements = bpmnTestcase.getFlow().getExclusiveFlowNodes();
		List<BPMNGenFlowElement> allElements = new ArrayList<>();
		allElements.addAll(flowElements);
		allElements.addAll(exclusiveFlowElements);
		
		allElements.stream() //
				.distinct() //
				.forEach(passCheckRelevant -> passCheckList.add(new PassCheck(passCheckRelevant, false))); //
	}

	private void resetPassCheckMap() {
		passCheckList.forEach(entry -> entry.setPassCheck(false)); //
	}

	private void markFragmentFlowElementsAsShouldHavePassed(List<BPMNGenFlowElement> fragment) {
		fragment.stream() //
				.map(flowElement -> flowElement.getId()) //
				.forEach(flowElementId -> markFlowElementAsShouldHavePassed(flowElementId)); //
	}

	private void markFlowElementAsShouldHavePassed(String flowElementId) {
		passCheckList.stream() //
				.filter(passCheck -> passCheck.getFlowElementId().equals(flowElementId)) //
				.forEach(passCheck -> passCheck.setPassCheck(true));
	}

	private IsPassingCheck mapToPassingCheck(String taskId, FlowElementType flowElementType, boolean shouldHavePassed) {
		if (shouldHavePassed) {
			return new PassedCheck(taskId, flowElementType);
		}

		return new NotPassedCheck(taskId, flowElementType);
	}

	private void optimizePassCheckOrder(List<BPMNGenFlowElement> flowElements) {
		passCheckList.sort(new PassCheckComparator(flowElements));
	}

	private HashMap<String, Object> getVariablesAndPop(List<TaskVariables> taskVariables, String task) {
		if(!taskVariables.isEmpty()) {
			TaskVariables firstTaskVariables = taskVariables.get(0);
			
			if(firstTaskVariables.getTask().equals(task)) {
				taskVariables.remove(0);
				return firstTaskVariables.getVariables();
			}
		}
		
		return null;
	}
	
	public class PassCheck {

		private BPMNGenFlowElement flowElement;
		private boolean passCheck;

		public PassCheck(BPMNGenFlowElement flowElement, boolean passCheck) {
			this.flowElement = flowElement;
			this.passCheck = passCheck;
		}

		public BPMNGenFlowElement getFlowElement() {
			return flowElement;
		}

		public void setFlowElement(BPMNGenFlowElement flowElement) {
			this.flowElement = flowElement;
		}

		public String getFlowElementId() {
			return flowElement.getId();
		}

		public boolean isPassCheck() {
			return passCheck;
		}

		public void setPassCheck(boolean passCheck) {
			this.passCheck = passCheck;
		}
	}

	public class PassCheckComparator implements Comparator<PassCheck> {

		private List<BPMNGenFlowElement> model;

		public PassCheckComparator(List<BPMNGenFlowElement> model) {
			this.model = model;
		}

		@Override
		public int compare(PassCheck left, PassCheck right) {
			return Integer.compare(getIndexOrMax(left.getFlowElement()), getIndexOrMax(right.getFlowElement()));
		}

		private int getIndexOrMax(BPMNGenFlowElement flowElement) {
			int index = model.indexOf(flowElement);

			return index != -1 ? index : Integer.MAX_VALUE;
		}
	}
}
